/*
 * New LZW packing program
 *
 * (c) 1992 Richard Kettlewell
 *
 * See lzw_txt.
 */
#include	<stdio.h>
#include	<stdlib.h>
#include	<assert.h>

#define MAXBITS			12				/* Maximum bits in a code (8 < MAXBITS < 16) */
#define HASHBITS		6				/* Bits in a hash code (5 or 6 best) */
#define CHARBITS		8				/* Bits in a character */

#define CODES			(1<<MAXBITS)	/* How many codes */
#define HASHES			(1<<HASHBITS)	/* Number of distinct hash codes */
#define CHARS			(1<<CHARBITS)	/* Number of distinct characters */

struct NODE {
	unsigned char character;			/* Last character in string */
	unsigned char first;				/* First character in string */
	short root;							/* Previous characters, or -1 */
	struct NODE *next;					/* Another node with same hash code */
};

int byte, count;						/* Used for bit io */
int inmax = CHARS-1;					/* Maximum inputable code */
int nodes = CHARS;						/* Nodes so far */
struct NODE *hashes[HASHES];			/* Hash table */
struct NODE table[CODES];				/* The big list! */

char _prog_name[] = "LZW";
char _version[] = "2.0";				/* Totally re-written since 1.0 */
char _copyright[] = "(c) 1992 RJK";
int _stack = 16000;

int hashcode(int, int);
int input_bit(FILE *);
int input_code(FILE *);
int pack(FILE *,FILE *);
int unpack(FILE *,FILE *);
void add_node(int, int);
void output_bit(int, FILE *);
void output_code(int, FILE *);
void output_flush(FILE *);
int output_string(int, FILE *);
void setup_table(void);
void usage(void);

/*
 * Entry point
 */
int main(int argc, char **argv)
{
	assert(sizeof (struct NODE) == 8);
	if(argc != 2) {
		usage();
		return -15;
	}
	switch(argv[1][0]) {
		case 'c': case 'C':
			return pack(stdin,stdout);
		case 'd': case 'D':
			return unpack(stdin,stdout);
		default:
			usage();
			return -15;
	}
}

/*
 * Tell user what to do
 */
void usage(void)
{
	fprintf(stderr,"Usage:-\n\n  LZW c|d <input >output\n\n");
}

/*
 * Given a character and a root, generate the appropriate hash code
 */
int hashcode(int root, int character)
{
	int hash = 0;
	int i;

	for(i = CHARBITS-(CHARBITS & 1); i ; i -= 2) {
		hash ^= character;
		character >>= 2;
	}
	for(i = MAXBITS-(MAXBITS & 1); i; i -= 2) {
		hash ^= root;
		root >>= 2;
	}
	return hash & (HASHES-1);
}

/*
 * Prepare the table
 */ 
void setup_table(void)
{
	int i;

	for(i = 0; i < CHARS; ++i) {
		table[i].character = table[i].first = i;
		table[i].root = -1;
	}
}

/*
 * Pack a file
 */
int pack(in,out)
FILE *in,*out;
{
	int code;
	int K;
	long length;

	setup_table();
	byte = 0;
	count = 1;
	while(EOF != getc(in))
		;
	if(ferror(in))
		return -16;
	length = ftell(in);
	rewind(in);
	/* Output the length in a portable way (fwrite is not portable for this) */
	fputc(length, out);
	fputc(length >>= 8, out);
	fputc(length >>= 8, out);
	fputc(length >>= 8, out);
	/* This implementation copes with zero length files */
	code = -1;
	while(EOF != (K = getc(in))) {
		/* Find if code + K is in table */
		if(code == -1)
			code = K;
		else {
			struct NODE *n;
			n = hashes[hashcode(code, K)];
			while(n && (n->root != code || n->character != K))
				n = n->next;
			if(n)
				code = n-table;			/* Heh heh! */
			else {
				output_code(code, out);
				add_node(code, K);
				code = K;
			}
		}
	}
	if(code != -1)
		output_code(code, out);
	output_flush(out);
	return ferror(out) ? -16 : 0;
}

/*
 * Create a new node if possible
 */
void add_node(int root, int character)
{
	int h;
	struct NODE *n;

	if(nodes < CODES) {
		n = &table[nodes++];
		h = hashcode(root, character);
		n->next = hashes[h];
		hashes[h] = n;
		n->character = character;
		n->first = table[root].first;
		n->root = root;
	}
}

/*
 * Output a code using the appropriate number of bits
 */
void output_code(int code, FILE *fp)
{
	int M = nodes - 1;

	do {
		output_bit(code & 1, fp);
		M >>= 1;
		code >>= 1;
	} while(M != 0);
}

/*
 * Output a single bit
 */ 
void output_bit(int bit, FILE *fp)
{
	if(bit)
		byte += count;
	count <<= 1;
	if(count == CHARS) {
		putc(byte, fp);
		count = 1;
		byte = 0;
	}
}

/*
 * Finish of bit output
 */ 
void output_flush(FILE *fp)
{
	if(count != 1) {
		putc(byte, fp);
		byte = 0;
		count = 1;
	}
}

/*
 * Unpack a file
 */ 
int unpack(in,out)
FILE *in,*out;
{
	int i, code, old, K;
	unsigned long length;

	count = CHARS>>1;
	setup_table();
	length = 0;
	for(i = 0; i < 32; i += 8)
		length += getc(in)<<i;
	/* We have to kludge zero length files here */
	if(!length)
		return 0;
	code = input_code(in);
	length -= output_string(code, out);
	while(length) {
		old = code;
		code = input_code(in);
		if(code < nodes) {
			length -= output_string(code, out);
			K = table[code].first;
		} else {
			K = table[old].first;
			length -= 1 + output_string(old, out);
			putc(K, out);
		}
		add_node(old, K);
	}
	return ferror(in) || ferror(out) ? -16 : 0;
}

/*
 * Read a code using the appropriate bit size
 */
int input_code(FILE *fp)
{
	int M = inmax;
	int C = 0;
	int B = 1;

	do {
		if(input_bit(fp))
			C += B;
		B += B;
		M >>= 1;
	} while(M != 0);
	++inmax;
	return C;
}

/*
 * Read a bit
 */
int input_bit(FILE *fp)
{
	count += count;
	if(count == CHARS) {
		byte = getc(fp);
		count = 1;
	}
	return byte & count ? 1 : 0;
}

/*
 * Output the string for a code
 * Return the number of characters written
 */
int output_string(int code, FILE *fp)
{
	int next;
	int count;

	if(-1 != (next = table[code].root))
		count = output_string(next, fp);
	else
		count = 0;
	putc(table[code].character, fp);
	return count+1;
}
